from typing import *

import ibis
from mypy_extensions import TypedDict
from typing_extensions import Literal

__all__ = ["bin"]

BinTransform = TypedDict(
    "BinTransform",
    {
        # Original field name
        "field": str,
        "type": Literal["bin"],
        # the  name of the columns to store the left and right side of the bin
        "as": Tuple[str, str],
        # The number of bins
        "maxbins": int,
        # field that we should get the extent from
        "extent": str,
    },
)


def bin(transform: BinTransform, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega bin transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/bin/
    """

    field = expr[transform["field"]]
    as_left, as_right = transform["as"]
    maxbins = transform["maxbins"]
    extent = expr[transform["extent"]]

    # Precompute min/max or else we get
    # "Expression 'xxx' is not being grouped"
    # errors
    min_ = extent.min().execute()
    max_ = extent.max().execute()

    # Cast these to floats to work around
    # https://github.com/ibis-project/ibis/issues/1934
    binwidth = (max_ - min_) / maxbins

    bin_ = (((field / _float(binwidth)) - _float(min_ / binwidth))).floor()
    left = (min_ + (bin_ * binwidth)).name(as_left)
    right = (((min_ + binwidth) + (bin_ * binwidth))).name(as_right)

    # raise ''
    # add the two new fields and remove the initial column
    return expr.mutate(
        [left, right]
        # + [c for c in expr.columns if c not in {transform["field"], as_left, as_right}]
    )


def _float(value) -> ibis.Expr:
    return ibis.literal(value, "float64").cast("float32")



"""
Functionality for server-side ibis transforms of vega charts.
"""
import json
import re
import typing

import altair
import altair.vegalite.v3.display
import opentracing

from .core import apply
from .globals import _expr_map
from .tracer import tracer

__all__ = ["query_target_func"]


def query_target_func(comm, msg):
    """
    Target function for actually evaluating the `queryibis` transform.
    """
    # These are the paramaters passed to the vega transform
    parameters: dict = msg["content"]["data"]

    name: str = parameters.pop("name")
    transforms: typing.Optional[str] = parameters.pop("transform", None)
    injected_span: object = parameters.pop("span")

    with tracer.start_span(
        "queryibis",
        references=[
            opentracing.child_of(
                (tracer.extract(opentracing.Format.TEXT_MAP, injected_span))
            )
        ],
    ) as span:
        if name not in _expr_map:
            raise ValueError(f"{name} is not an expression known to us!")
        expr = _expr_map[name]
        span.log_kv({"sql:initial": expr.compile(), "transforms": transforms})
        if transforms:
            # Replace all string instances of data references with value in schema
            for k, v in parameters.items():
                # All data items are added to parameters as `:<data name>`.
                # They also should  be in the `data` paramater, but you have to call
                # this with a tuple which I am not sure where to get from
                # https://github.com/vega/vega/blob/65fe7cb2485be90e16298d9dff87bf56045afb8d/packages/vega-transforms/src/Filter.js#L48
                if not k.startswith(":"):
                    continue
                k = k[1:]
                res = json.dumps(v)
                for t in transforms:
                    if t["type"] == "filter" or t["type"] == "formula":
                        t["expr"] = _patch_vegaexpr(t["expr"], k, res)
            try:
                expr = apply(expr, transforms)
            except Exception as e:
                raise ValueError(
                    f"Failed to convert {transforms} with error message message '{e}'"
                )
        with tracer.start_span("ibis:execute", child_of=span) as execute_span:
            execute_span.log_kv({"sql": expr.compile()})
            data = expr.execute()
        comm.send(altair.to_values(data)["values"])


def _patch_vegaexpr(expr: str, name: str, value: str) -> str:
    quote = "(['\"])"
    expr = re.sub(f"data\({quote}{name}{quote}\)", value, expr)
    expr = re.sub(
        f"vlSelectionTest\({quote}{name}{quote}", f"vlSelectionTest({value}", expr
    )
    return expr



import warnings

import altair
import ibis
import pandas

__all__ = ["monkeypatch_altair"]


def monkeypatch_altair():
    """
    Needed until https://github.com/altair-viz/altair/issues/843 is fixed to let Altair
    handle ibis inputs
    """
    original_chart_init = altair.Chart.__init__
    original_chart_to_dict = altair.Chart.to_dict

    def updated_chart_init(self, data=None, *args, **kwargs):
        """
        If user passes in a Ibis expression, create an empty dataframe with
        those types and set the `ibis` attribute to the original ibis expression.
        """
        if data is not None and isinstance(data, ibis.Expr):
            expr = data
            data = empty_dataframe(expr)
            data.ibis = expr

        return original_chart_init(self, data=data, *args, **kwargs)

    def updated_chart_to_dict(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            spec = original_chart_to_dict(self, *args, **kwargs)
        return spec

    altair.Chart.__init__ = updated_chart_init
    altair.Chart.to_dict = updated_chart_to_dict


def empty_dataframe(expr: ibis.Expr) -> pandas.DataFrame:
    """
    Creates an empty DF for a ibis expression, based on the schema

    https://github.com/ibis-project/ibis/issues/1676#issuecomment-441472528
    """
    return expr.schema().apply_to(pandas.DataFrame(columns=expr.columns))





import os
import jaeger_client

__all__ = ["tracer"]


jaeger_config = jaeger_client.Config(
    config={"sampler": {"type": "const", "param": 1}, "logging": True},
    service_name=os.environ.get("JAEGER_SERVICE_NAME", "kernel"),
    validate=True,
)

tracer = jaeger_config.initialize_tracer()



import copy
import re
import typing

import opentracing

from .globals import DATA_NAME_PREFIX, _expr_map
from .tracer import tracer

__all__ = ["compiler_target_function"]

# An empty vega spec to send when we get invalid data.
EMPTY_VEGA = {
    "$schema": "https://vega.github.io/schema/vega/v5.json",
    "description": "An empty vega v5 spec",
    "width": 500,
    "height": 200,
    "padding": 5,
    "autosize": "pad",
    "signals": [],
    "data": [],
    "scales": [],
    "projections": [],
    "axes": [],
    "legends": [],
    "marks": [],
}


def compiler_target_function(comm, msg):
    """
    A function that takes a vega spec and converts its vega-transforms
    to ibis transforms, which will later be able to lazily evaluate
    upon user interactions.
    """
    data = msg["content"]["data"]
    spec = data["spec"]
    injected_span = data["span"]
    root_span = data["rootSpan"]

    root_ref = opentracing.child_of(
        (tracer.extract(opentracing.Format.TEXT_MAP, injected_span))
    )

    with tracer.start_span("compile-vega", references=root_ref) as span:
        span.log_kv({"vega-spec:initial": spec})

        try:
            with tracer.start_span("transform-vega", child_of=span) as transform_span:
                updated_spec = _transform(spec, root_span)
                transform_span.log_kv({"vega-spec:transformed": updated_spec})

            comm.send(updated_spec)
        except ValueError as e:
            # If there was an error transforming the spec, which can happen
            # if we don't support all the required transforms, or if
            # the spec references an old, unavailable ibis expression,
            # then send an empty vega spec.
            comm.send(EMPTY_VEGA)
            raise e


def _transform(
    spec: typing.Dict[str, typing.Any], root_span: object
) -> typing.Dict[str, typing.Any]:
    """
    Transform a vega spec into one that uses the `queryibis` transform
    in the place of vega transforms.
    """
    new = copy.deepcopy(spec)  # make a copy of the spec
    _transforms = {}  # Store references to ibis query transforms
    _root_expressions = {}  # Keep track of the sources backed in ibis expressions

    # We iteratively pass through the data, looking
    # for named ibis data sources and data sources that
    # reference our named sources. When we can pass through
    # the entire set of data without encountering one,
    # then we are done.
    done = False
    while not done:
        for data in new["data"]:
            # First check for named data which matches an initial
            # ibis expression passed in directly via altair.
            name = data.get("name", "")
            if _is_ibis(name) and name not in _root_expressions:
                key = _retrieve_expr_key(name)
                new_transform = {"type": "queryibis", "name": key, "span": root_span}
                # If the named data has transforms, set them
                # in the ibis transform, and keep a reference
                # to them in case we need to incorporate them
                # into downstream data attributes.
                old_transform = data.get("transform", None)
                if old_transform:
                    _transforms[name] = old_transform
                    new_transform["transform"] = old_transform
                data["transform"] = [new_transform]
                _root_expressions[name] = name
                break

            # Next check to see if the data sources an upstream data set.
            # If it is one of ours, transform that as well.
            source = data.get("source", "")
            if source and source in _root_expressions:
                del data["source"]
                source_transforms = _transforms.get(source, [])
                old_transforms = data.get("transform", [])

                # Rename "signal" to "signal_" because if vega
                # sees a "signal" key on an object it will try to resolve it, instead of passing
                # it into the transform
                for t in old_transforms:
                    if "signal" in t:
                        t["signal_"] = t["signal"]
                        del t["signal"]

                new_transforms = source_transforms + old_transforms
                data["transform"] = [
                    {
                        "type": "queryibis",
                        "name": _retrieve_expr_key(_root_expressions[source]),
                        "span": root_span,
                        "data": "{"
                        + ", ".join(
                            f"{field}: data('{field}')"
                            for field in _extract_used_data(new_transforms)
                        )
                        + "}",
                        "transform": new_transforms,
                    }
                ]
                _root_expressions[name] = _root_expressions[source]
                _transforms[name] = new_transforms
                break
        # If we make it through the data sources without a `break`
        # statement, there are no more data sources to transform,
        # so we are done.
        else:
            done = True

    return _cleanup_spec(new)


def _extract_used_data(transforms) -> typing.Set[str]:
    """
    Given a list of transforms, returns a set of the data fields they depend on.
    """
    return {m.group(1) for m in re.finditer(r'data\("(.+?)"\)', str(transforms))}


assert _extract_used_data(
    [
        {
            "type": "filter",
            "expr": '!(length(data("Filter_store"))) || (vlSelectionTest("Filter_store", datum))',
        }
    ]
) == {"Filter_store"}


def _is_ibis(name: str) -> bool:
    """
    Test whether a vega data name refers to one of the ibis expressions.
    """
    return name.startswith(DATA_NAME_PREFIX)


def _retrieve_expr_key(name: str) -> typing.Optional[str]:
    if not name.startswith(DATA_NAME_PREFIX):
        return None
    key = name[len(DATA_NAME_PREFIX) :]
    if key not in _expr_map:
        raise ValueError(f"Unrecognized ibis data name {name}")
    return key


def _cleanup_spec(spec):
    """
    Goes through the spec and removes data sources that are not referenced
    anywhere else in the spec. Does this by turning the spec into a string
    and seeing if the name of the data is in the string.
    """

    nonreferenced_data = []
    for data in spec["data"]:
        name = data["name"]
        # create a vesion of the spec where this data is removed
        without_this_data = copy.deepcopy(spec)
        without_this_data["data"].remove(data)
        has_reference = name in str(without_this_data)
        if not has_reference:
            nonreferenced_data.append(data)

    new = copy.deepcopy(spec)
    new["data"] = [data for data in new["data"] if data not in nonreferenced_data]
    return new




import altair
import opentracing
import typing

from .globals import get_active_span, get_fallback, set_active_span
from .tracer import tracer

__all__ = ["altair_renderer"]

# New Vega Lite renderer mimetype which can process ibis expressions in names
MIMETYPE = "application/vnd.vega.ibis.v5+json"


def altair_renderer(spec):
    """
    An altair renderer that serves our custom mimetype with ibis support.
    """
    if get_fallback():
        return altair.vegalite.v3.display.default_renderer(spec)

    active_span = get_active_span()
    assert active_span
    injected_span: typing.Dict = {}
    tracer.inject(active_span, opentracing.Format.TEXT_MAP, injected_span)
    active_span.log_kv({"vega-lite:initial": spec})
    active_span.finish()
    set_active_span(None)

    return {MIMETYPE: {"spec": spec, "span": injected_span}}



import altair
import pandas
from opentracing import tags

from .globals import (
    DATA_NAME_PREFIX,
    _expr_map,
    get_active_span,
    get_fallback,
    set_active_span,
)
from .tracer import tracer

__all__ = ["altair_data_transformer"]


def altair_data_transformer(data):
    """
    turn a pandas DF with the Ibis query that made it attached to it into
    a valid Vega Lite data dict. Since this has to be JSON serializiable
    (because of how Altair is set up), we create a unique name and
    save the ibis expression globally with that name so we can pick it up later.
    """
    assert isinstance(data, pandas.DataFrame)
    expr = data.ibis

    if get_fallback():
        return altair.default_data_transformer(expr.limit(1000).execute())
    # Start a span during the first data transform
    if not get_active_span():
        set_active_span(tracer.start_span("altair", tags={tags.SERVICE: "kernel"}))

    get_active_span().log_kv({"sql:initial": expr.compile()})

    h = str(hash(expr))
    name = f"{DATA_NAME_PREFIX}{h}"
    _expr_map[h] = expr
    return {"name": name}




"""Ibis-Vega Transform

This module provides a Python implementation of Vega transforms,
evaluated with ibis expressions.
The main function is the ``ibis_vega_transform.apply()`` function.
"""
import altair
import IPython

from .altair_data_transformer import altair_data_transformer
from .altair_monkeypatch import monkeypatch_altair
from .altair_renderer import altair_renderer
from .compiler import compiler_target_function
from .core import apply
from .query import query_target_func
from .globals import set_fallback, _expr_map

__all__ = ["set_fallback", "_expr_map", "apply"]


##
# Set global state
##

ipython = IPython.get_ipython()
kernel = getattr(ipython, "kernel", None) if ipython else None

if kernel:
    kernel.comm_manager.register_target("queryibis", query_target_func)
    kernel.comm_manager.register_target(
        "ibis-vega-transform:compiler", compiler_target_function
    )


monkeypatch_altair()

altair.renderers.register("ibis", altair_renderer)
altair.renderers.enable("ibis")

altair.data_transformers.register("ibis", altair_data_transformer)
altair.data_transformers.enable("ibis")



import typing
import ibis
import opentracing

__all__ = [
    "_expr_map",
    "DATA_NAME_PREFIX",
    "get_fallback",
    "set_fallback",
    "get_active_span",
    "set_active_span",
]

_expr_map: typing.Dict[str, ibis.Expr] = {}

DATA_NAME_PREFIX = "ibis:"


# Whether to fallback to getting the dataset as a pandas dataframe
# and rendering it manually, that way.
FALLBACK = False


def set_fallback(fallback: bool) -> None:
    global FALLBACK
    FALLBACK = fallback


def get_fallback() -> bool:
    return FALLBACK


active_span: typing.Optional[opentracing.Span] = None


def set_active_span(scan: typing.Optional[opentracing.Span]) -> None:
    global active_span
    active_span = scan


def get_active_span() -> typing.Optional[opentracing.Span]:
    return active_span



import ibis

from ibis_vega_transform.util import promote_list


def collect(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega collect transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/collect/

    Parameters
    ----------
    transform: dict
        A JSON-able dictionary representing the vega transform.
    expr: ibis.Expr
        The expression to which to apply the transform.

    Returns
    -------
    transformed_expr: the transformed expression
    """
    fields = promote_list(transform["sort"]["field"])
    orders = promote_list(transform["sort"].get("order", ["ascending"] * len(fields)))
    assert len(fields) == len(orders)

    rules = [
        (field, (True if order == "ascending" else False))
        for field, order in zip(fields, orders)
    ]
    return expr.sort_by(rules)



"""
Evaluate vega expressions using ibis
"""
import datetime as dt
import functools
import math
import operator
import random
import sys
from typing import *

import altair_transform.utils._evaljs
import ibis
import ibis.expr.types as it
from altair_transform.utils import evaljs
from mypy_extensions import TypedDict
from typing_extensions import Literal

# Monkey patch altair_transform so that boolean operators  work on ibis expression


def not_operator(value):
    if isinstance(value, ibis.expr.types.ValueExpr):
        return ~value
    return not value


def and_operator(l, r):
    if isinstance(l, ibis.expr.types.ValueExpr):
        if isinstance(r, ibis.expr.types.ValueExpr):
            return l & r
        return r and l
    return l and r


def or_operator(l, r):
    if isinstance(l, ibis.expr.types.ValueExpr):
        if isinstance(r, ibis.expr.types.ValueExpr):
            return l & r
        return r or l
    return l or r


# We also want to monkey patch `x == null` so that it resolves to `(x IS NULL)`, because
# the null literal is not available for omnisci.


def equal_operator(l, r):
    if ibis.null().equals(l):
        return r.isnull()
    if ibis.null().equals(r):
        return l.isnull()
    return l == r


# and the same for not equal
def not_equal_operator(l, r):
    return not_operator(equal_operator(l, r))


altair_transform.utils._evaljs.UNARY_OPERATORS["!"] = not_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["&&"] = and_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["||"] = or_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["=="] = equal_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["==="] = equal_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["!="] = not_equal_operator
altair_transform.utils._evaljs.BINARY_OPERATORS["!=="] = not_equal_operator


def eval_vegajs(expression: str, datum: ibis.Expr = None) -> ibis.Expr:
    """Evaluate a vega expression"""
    namespace = {"datum": datum} if datum is not None else {}
    namespace.update(VEGAJS_NAMESPACE)
    return evaljs(expression, namespace)


# Type Coercion Functions
def isArray(value: Any) -> bool:
    """Returns true if value is an array, false otherwise."""
    return False


def isBoolean(value: Any) -> bool:
    """Returns true if value is a boolean (true or false), false otherwise."""
    return isinstance(value, it.BooleanValue)


def isDate(value: Any) -> bool:
    """Returns true if value is a Date object, false otherwise.
    """
    return isinstance(value, it.TemporalValue)


def isNumber(value: Any) -> bool:
    """Returns true if value is a number, false otherwise.

    NaN and Infinity are considered numbers.
    """
    return isinstance(value, it.NumericValue)


def isObject(value: Any) -> bool:
    """Returns true if value is an object, false otherwise.

    Following JavaScript typeof convention, null values are considered objects.
    """
    return value is None or isinstance(value, it.MapValue)


def isRegExp(value: Any) -> bool:
    """
    Returns true if value is a RegExp (regular expression)
    object, false otherwise.
    """
    return False


def isString(value: Any) -> bool:
    """Returns true if value is a string, false otherwise."""
    return isinstance(value, it.StringValue)


# Type Coercion Functions
def toBoolean(value: Any) -> bool:
    """
    Coerces the input value to a boolean.
    Null values and empty strings are mapped to null.
    """
    return value.cast("boolean")


def toDate(value: Any) -> Optional[dt.datetime]:
    """
    Coerces the input value to a Date instance.
    Null values and empty strings are mapped to null.
    If an optional parser function is provided, it is used to
    perform date parsing, otherwise Date.parse is used.
    """
    return value.cast("timestamp")


def toNumber(value: Any) -> Optional[float]:
    """
    Coerces the input value to a number.
    Null values and empty strings are mapped to null.
    """
    return value.cast("double")


def toString(value: Any) -> Optional[str]:
    """
    Coerces the input value to a string.
    Null values and empty strings are mapped to null.
    """
    return value.cast("string")


# Date/Time Functions
def now() -> float:
    """Returns the timestamp for the current time."""
    return ibis.now()


def z(v) -> bool:
    """
    Returns whether the value is equal to a literal 0
    """
    return not isinstance(v, ibis.expr.types.Expr) and v == 0


def datetime(y, m, d, h, mi, s, ms):
    """
    This should use https://github.com/ibis-project/ibis/issues/386 if it gets implemented

    Currently we only support use cases that can be turned into a truncate, which is:

        datetime(year(col), 0, 0, 0, 0, 0)
        or
        datetime(year(col), month(col), 0, 0, 0, 0)
        etc.
    """
    if not hasattr(y, "_arg"):
        raise NotImplementedError()
    if not isinstance(y._arg, ibis.expr.operations.ExtractYear):
        raise NotImplementedError()
    date_column = y._arg.arg

    if z(m):
        if not z(d) or not z(h) or not z(mi) or not z(s) or not z(ms):
            raise NotImplementedError()
        return date_column.truncate("year")
    if not m.equals(month(date_column)):
        raise NotImplementedError()

    if z(d):
        if not z(h) or not z(mi) or not z(s) or not z(ms):
            raise NotImplementedError()
        return date_column.truncate("month")
    if not d.equals(date(date_column)):
        raise NotImplementedError(f"{d}, {day(date_column)}")

    if z(h):
        if not z(mi) or not z(s) or not z(ms):
            raise NotImplementedError()
        return date_column.truncate("day")
    if not h.equals(hours(date_column)):
        raise NotImplementedError()

    if z(mi):
        if not z(s) or not z(ms):
            raise NotImplementedError()
        return date_column.truncate("hour")
    if not mi.equals(minutes(date_column)):
        raise NotImplementedError()

    if z(s):
        if not z(ms):
            raise NotImplementedError()
        return date_column.truncate("minute")
    if not s.equals(seconds(date_column)):
        raise NotImplementedError()

    if z(ms):
        return date_column.truncate("second")
    if not ms.equals(milliseconds(date_column)):
        raise NotImplementedError()
    return date_column


def date(datetime):
    """
    Returns the day of the month for the given datetime value, in local time.
    """
    return datetime.day()


def day(datetime):
    """
    Returns the day of the week for the given datetime value, in local time.
    """
    return datetime.day_of_week.index()


def year(datetime):
    """Returns the year for the given datetime value, in local time."""
    return datetime.year()


def quarter(datetime):
    """
    Returns the quarter of the year (0-3) for the given datetime value,
    in local time.
    """
    return (datetime.month() - 1) // 3


def month(datetime):
    """
    Returns the (zero-based) month for the given datetime value, in local time.
    """
    return datetime.month() - 1


def hours(datetime):
    """
    Returns the hours component for the given datetime value, in local time.
    """
    return datetime.hour()


def minutes(datetime):
    """
    Returns the minutes component for the given datetime value, in local time.
    """
    return datetime.minute()


def seconds(datetime):
    """
    Returns the seconds component for the given datetime value, in local time.
    """
    return datetime.second()


def milliseconds(datetime) -> float:
    """
    Returns the milliseconds component for the given datetime value,
    in local time.
    """
    return datetime.millisecond()


def time(datetime) -> float:
    """Returns the epoch-based timestamp for the given datetime value."""
    raise NotImplementedError("time()")


def timezoneoffset(datetime):
    # TODO: use tzlocal?
    raise NotImplementedError("timezoneoffset()")


def utc(year, month, day=0, hour=0, min=0, sec=0, millisec=0):
    """
    Returns a timestamp for the given UTC date.
    The month is 0-based, such that 1 represents February.
    """
    raise NotImplementedError("utc()")
    # return dt.datetime(
    #     int(year),
    #     int(month) + 1,
    #     int(day),
    #     int(hour),
    #     int(min),
    #     int(sec),
    #     int(millisec * 1000),
    #     tzinfo=pytz.UTC,
    # )


FieldDict = TypedDict(
    "FieldDict",
    {
        "field": str,
        # identifies whether tuples in the dataset enumerate values for the
        # field, or specify a continuous range.
        "type": Literal["E", "R", "R-E", "R-LE", "R-RE"],
    },
)

SelectionDict = TypedDict("SelectionDict", {"fields": List[FieldDict], "values": List})


JS_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def _test_single_point(expr: ibis.Expr, field: FieldDict, value: Any) -> ibis.Expr:
    column = expr[field["field"]]
    tp = field["type"]

    # SPECIAL CASE
    # Workaround for case where we are selecting a range of dates
    # for some reason this appears as type E even though it should be type R
    if isinstance(column, ibis.expr.types.TemporalValue) and tp == "E":
        tp = "R"
        value = [dt.datetime.strptime(v, JS_DATETIME_FORMAT) for v in value]

    if tp == "E":
        if isinstance(value, list):
            return column.isin(value)
        return column == value
    lower, upper = value
    if tp == "R":
        return (lower <= column) & (column <= upper)
    if tp == "R-RE":
        return (lower <= column) & (column < upper)
    if tp == "R-LE":
        return (lower < column) & (column <= upper)
    raise NotImplementedError(f"don't recoognize {tp}")


def _test_point(expr: ibis.Expr, entry: SelectionDict) -> ibis.Expr:
    """
    Translated from
    https://github.com/vega/vega/blob/353a4097a5c726ec6b5b1df71722976d246c6cd7/packages/vega-selections/src/selectionTest.js#L12-L48
    and
    https://github.com/vega/vega/blob/master/packages/vega-util/src/inrange.js
    """
    return functools.reduce(
        operator.and_,
        (
            _test_single_point(expr, field, value)
            for field, value in zip(entry["fields"], entry["values"])
        ),
    )


def vlSelectionTest(
    filters: List[SelectionDict],
    expr: ibis.Expr,
    op: Literal["union", "intersect"] = "union",
) -> ibis.Expr:
    """
    Instead of passing in the data name as the first arg, we pass in the actual data, like:

    >>>  [{'fields': [{'type': 'E', 'field': 'c'}], 'values': ['second']}]

    Translated from:

    https://github.com/vega/vega/blob/353a4097a5c726ec6b5b1df71722976d246c6cd7/packages/vega-selections/src/selectionTest.js#L50-L63
    """
    if not filters:
        return expr
    return functools.reduce(
        {"union": operator.or_, "intersect": operator.and_}[op],
        (_test_point(expr, f) for f in filters),
    )


# From https://vega.github.io/vega/docs/expressions/
VEGAJS_NAMESPACE: Dict[str, Any] = {
    # Constants
    "null": ibis.null(),
    "NaN": ibis.NA,
    "E": math.e,
    "LN2": math.log(2),
    "LN10": math.log(10),
    "LOG2E": math.log2(math.e),
    "LOG10E": math.log10(math.e),
    "MAX_VALUE": sys.float_info.max,
    "MIN_VALUE": sys.float_info.min,
    "PI": math.pi,
    "SQRT1_2": math.sqrt(0.5),
    "SQRT2": math.sqrt(2),
    # Type Checking
    "isArray": isArray,
    "isBoolean": isBoolean,
    "isDate": isDate,
    "isNumber": isNumber,
    "isObject": isObject,
    "isRegExp": isRegExp,
    "isString": isString,
    # Type Coercion
    "toBoolean": toBoolean,
    "toDate": toDate,
    "toNumber": toNumber,
    "toString": toString,
    # Control Flow Functions
    "if": lambda test, true_expr, false_expr: ibis.ifelse(test, true_expr, false_expr),
    # Math Functions
    "isNaN": lambda x: x.isnull(),
    "abs": ibis.expr.api.abs,
    "acos": ibis.expr.api.acos,
    "asin": ibis.expr.api.asin,
    "atan": ibis.expr.api.atan,
    "atan2": ibis.expr.api.atan2,
    "ceil": ibis.expr.api.ceil,
    "cos": ibis.expr.api.cos,
    "exp": ibis.expr.api.exp,
    "floor": ibis.expr.api.floor,
    "log": ibis.expr.api.ln,
    "max": ibis.expr.api.max,
    "min": ibis.expr.api.min,
    "pow": ibis.expr.api.pow,
    "random": random.random,
    "round": ibis.expr.api.round,
    "sin": ibis.expr.api.sin,
    "sqrt": ibis.expr.api.sqrt,
    "tan": ibis.expr.api.tan,
    # Date/Time Functions
    "now": now,
    "datetime": datetime,
    "date": date,
    "day": day,
    "year": year,
    "quarter": quarter,
    "month": month,
    "hours": hours,
    "minutes": minutes,
    "seconds": seconds,
    "milliseconds": milliseconds,
    "time": time,
    "timezoneoffset": timezoneoffset,
    "utc": utc,
    "length": len,
    "vlSelectionTest": vlSelectionTest
    # TODOs:
    # Remaining Date/Time Functions
    # Array Functions
    # String Functions
    # Object Functions
    # Formatting Functions
    # RegExp Functions
}



from typing import *

import ibis
import ibis.expr.types as it


def aggregate(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega aggregate transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/aggregate/

    Parameters
    ----------
    transform: dict
        A JSON-able dictionary representing the vega transform.
    expr: ibis.Expr
        The expression to which to apply the transform.

    Returns
    -------
    transformed_expr: the transformed expression
    """
    groupby = transform["groupby"]
    # It's undocumented, but an undefined "ops" value defaults to ["count"]'
    # https://github.com/vega/vega/blob/4d10f9da0df0833c90ff259bbd0960f7cb05e3bf/packages/vega-transforms/src/Aggregate.js#L159-L161
    ops = transform.get("ops", ["count"])
    fields = transform.get("fields", [None])
    as_ = transform.get("as", [None])

    expr = expr.group_by(groupby).aggregate(
        [
            _aggregate(expr, field, op, as__)
            for (field, op, as__) in zip(fields, ops, as_)
        ]
    )
    return expr


def _aggregate(
    expr: ibis.Expr, field: str, op: str, name: Optional[str] = None
) -> ibis.Expr:
    """
    Apply an aggregation operation to an expression.

    Parameters
    ----------
    expr: ibis.Expr
        The expr to which to apply the aggregation operation.
    field: str
        The name of the field to aggregate.
    op: str
        The name of the aggregation operation, given by
        https://vega.github.io/vega/docs/transforms/aggregate/#ops
        Not all operations are implemented here.
    name: Optional[str]
        A name for the new aggregated value.
    """
    expr = expr[field] if field else expr
    operation = _translate_op(op)
    if not operation:
        raise ValueError(f"Unsupported op {op}")
    expr = operation(expr)
    return expr.name(name) if name else expr


def _translate_op(op: str):
    """
    Map a vega op (https://vega.github.io/vega/docs/transforms/aggregate/#ops)
    to an ibis expression operation.
    """
    return {
        "count": ibis.expr.api.count,
        "distinct": ibis.expr.api.distinct,
        "sum": ibis.expr.api.sum,
        "mean": ibis.expr.api.mean,
        "average": ibis.expr.api.mean,
        "variance": ibis.expr.api.variance,
        "stdev": ibis.expr.api.std,
        "median": ibis.expr.api.approx_median,
        "min": ibis.expr.api.min,
        "max": ibis.expr.api.max,
    }.get(op)



import ibis
from ibis_vega_transform.vegaexpr import eval_vegajs


def filter(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega filter transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/filter/

    This transform evaluates a vega expression, which is not fully
    implemented. Not every expression will work.

    Parameters
    ----------
    transform: dict
        A JSON-able dictionary representing the vega transform.
    expr: ibis.Expr
        The expression to which to apply the transform.

    Returns
    -------
    transformed_expr: the transformed expression
    """
    calc = transform["expr"]
    test = eval_vegajs(calc, expr)
    if test is True:
        return expr
    return expr[test]



import ibis
from ibis_vega_transform.vegaexpr import eval_vegajs


def formula(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega formula transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/formula/

    This transform evaluates a vega expression, which is not fully
    implemented. Not every expression will work.

    Parameters
    ----------
    transform: dict
        A JSON-able dictionary representing the vega transform.
    expr: ibis.Expr
        The expression to which to apply the transform.

    Returns
    -------
    transformed_expr: the transformed expression
    """
    col = transform["as"]
    calc = transform["expr"]
    new_col = eval_vegajs(calc, expr).name(col)
    return expr.mutate(new_col)



from typing import Any

import ibis

from . import transforms
from .util import promote_list

__all__ = ["apply"]


def apply(expr: ibis.Expr, transforms: Any) -> ibis.Expr:
    """Apply transform or transforms to the expression.

    Parameters
    ----------
    expr: ibis.Expr
    transform: list
        A transform specification or list of transform specifications.
        Each specification must be valid according to Vega's transform
        schema.

    Returns
    -------
    expr_transformed : ibis.expr
        The transformed dataframe.
    """
    if transforms is None:
        return expr
    transforms = promote_list(transforms)

    # First traverse list of transforms, and find any that create bins
    # The resulting bin fields, we create as the source fields,
    # Because, for some reason, the filter happens before
    # the binning, but it refers to the field created by the binning
    # See the signals in https://vega.github.io/editor/#/gist/9c7d4dee819450e59cf7381f4d47fee0/example.vl.json
    # as an example
    # TODO: Bring this up with Dominik and see why this is
    for t in transforms:
        if t["type"] == "bin":
            expr = expr.mutate(expr[t["field"]].name(t["as"][0]))

    # Have extra processing for extents that create signals
    # can probably remove once https://github.com/vega/vega-lite/issues/5320 is fixed.
    signal_mapping = {}

    for t in transforms:
        if t["type"] == "extent":
            assert {"field", "signal_", "type"} == t.keys()
            signal_mapping[t["signal_"]] = t["field"]
            continue
        # Change binning that reference  signal extent with actual value
        if "extent" in t and "signal" in t["extent"]:
            t["extent"] = signal_mapping.pop(t["extent"]["signal"])
        expr = _delegate_transform(t, expr)
    return expr


def _delegate_transform(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega transform to an ibis expression.

    Applying this function iteratively to an expression allows the user
    to build up a compound expression out of many vega transforms.

    If a particular transform is not implemented, raises an error.

    Parameters
    ----------
    transform: dict
        A JSON-able representation of a vega transform.
    expr: ibis.Expr
        An expression to transform.

    Returns
    -------
    expr_transformed: The original expression with the additional transform.
    """
    t = getattr(transforms, transform["type"])
    if t is not None:
        return t(transform, expr)
    else:
        raise NotImplementedError(f"Transform of type {t} is not implemented")



import os
import setuptools


def read(path, encoding="utf-8"):
    path = os.path.join(os.path.dirname(__file__), path)
    return open(path, encoding=encoding).read() if os.path.exists(path) else ""


setuptools.setup(
    name="ibis-vega-transform",
    version="2.0.1",
    url="https://github.com/Quansight/ibis-vega-transform",
    author="Ian Rose and Saul Shanabrook",
    author_email="ian.rose@quansight.com",
    license="Apache-2.0 license",
    description="Evaluate Vega transforms using Ibis expressions",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "jupyterlab",
        "altair-transform",
        "altair",
        "vega_datasets",
        "ibis-framework",
        "mypy_extensions",
        "typing_extensions",
        "jaeger-client",
        "jupyter_jaeger",
    ],
    extras_require={"dev": ["black"]},
    python_requires=">=3.6",
    include_package_data=True,
)



from typing import *

V = TypeVar("V")


def promote_list(val: Union[V, List[V]]) -> List[V]:
    """
    Maybe wrap a value in a list. If it is already a list, does nothing.
    """
    if not isinstance(val, list):
        val = [val]
    return val
